/*****************************************************************************
*
* Copyright (C) 2003 Atmel Corporation
*
* File              : USI_UART.c
* Compiler          : IAR EWAAVR 2.28a
* Created           : 18.07.2002 by JLL
* Modified          : 02-10-2003 by LTA
*
* Support mail      : avr@atmel.com
*
* Supported devices : ATtiny26
*
* Application Note  : AVR307 - Half duplex UART using the USI Interface
*
* Description       : Functions for USI_UART_receiver and USI_UART_transmitter.
*                     Uses Pin Change Interrupt to detect incomming signals.
*
*
****************************************************************************/
    
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "USI_UART_config.h"
#include "usi_spi_driver.h"

//********** USI UART Defines **********//

#define DATA_BITS                 8
#define START_BIT                 1
#define STOP_BIT                  1
#define HALF_FRAME                5

#define USI_COUNTER_MAX_COUNT     16
#define USI_COUNTER_SEED_TRANSMIT (USI_COUNTER_MAX_COUNT - HALF_FRAME)
#define INTERRUPT_STARTUP_DELAY   (0x11 / TIMER_PRESCALER)
// kein moderner AVR triggert USI per OVF:
// #define TIMER0_SEED               (256 - ( (SYSTEM_CLOCK / BAUDRATE) / TIMER_PRESCALER ))

#define FUDGE 5 // empirischer Mogelfaktor 5%
#define TIMER1_COMPC_MATCH				  ((((SYSTEM_CLOCK / BAUDRATE) / TIMER_PRESCALER ) + (FUDGE * SYSTEM_CLOCK / BAUDRATE) / TIMER_PRESCALER / 100))

#if (256 - (TIMER1_COMPC_MATCH * 1/2) < TIMER1_COMPC_MATCH )
#define INITIAL_TIMER1_SEED       (TIMER1_COMPC_MATCH * 1/2) // für halbe Bitlänge, um Mitte des StartBits zu treffen
#define USI_COUNTER_SEED_RECEIVE  ( USI_COUNTER_MAX_COUNT - (START_BIT + DATA_BITS) )
#else
#define INITIAL_TIMER1_SEED       (256-TIMER1_COMPC_MATCH * 1/2) // für 3* halbe Bitlänge, um Mitte des ersten DatenBits zu treffen
#define USI_COUNTER_SEED_RECEIVE  ( USI_COUNTER_MAX_COUNT - ( DATA_BITS) )
#endif

#if TIMER1_COMPC_MATCH == 109

#endif

#define UART_RX_BUFFER_MASK ( UART_RX_BUFFER_SIZE - 1 )
#if ( UART_RX_BUFFER_SIZE & UART_RX_BUFFER_MASK )
    #error RX buffer size is not a power of 2
#endif

#define UART_TX_BUFFER_MASK ( UART_TX_BUFFER_SIZE - 1 )
#if ( UART_TX_BUFFER_SIZE & UART_TX_BUFFER_MASK )
    #error TX buffer size is not a power of 2
#endif

/* General defines */
#define TRUE                      1
#define FALSE                     0

//********** Static Variables **********//

register uint8_t USI_UART_TxData asm("r3");
static char UART_RxBuf[UART_RX_BUFFER_SIZE];  // UART buffers. Size is definable in the header file.
static volatile uint8_t UART_RxHead;
static volatile uint8_t UART_RxTail;
static char UART_TxBuf[UART_TX_BUFFER_SIZE];
static volatile uint8_t UART_TxHead;
static volatile uint8_t UART_TxTail;

 static volatile union USI_UART_status                           // Status byte holding flags.
{
    uint8_t status;
    struct
    {
        uint8_t ongoing_Transmission_From_Buffer:1;
        uint8_t ongoing_Transmission_Of_Package:1;
        uint8_t ongoing_Reception_Of_Package:1;
        uint8_t reception_Buffer_Overflow:1;
        uint8_t flag4:1;
        uint8_t flag5:1;
        uint8_t flag6:1;
        uint8_t flag7:1;
    }__attribute__ ((packed));
} USI_UART_status = {0};


//********** USI_UART functions **********//

// Reverses the order of bits in a byte.
// I.e. MSB is swapped with LSB, etc.
__attribute__ ((const)) unsigned char Bit_Reverse( unsigned char x )
{
    x = ((x >> 1) & 0x55) | ((x << 1) & 0xaa);
    x = ((x >> 2) & 0x33) | ((x << 2) & 0xcc);
    x = ((x >> 4) & 0x0f) | ((x << 4) & 0xf0);
    return x;    
}

// Flush the UART buffers.
void USI_UART_Flush_Buffers( void )  
{  
    UART_RxTail = 0;
    UART_RxHead = 0;
    UART_TxTail = 0;
    UART_TxHead = 0;
    return;
}

/*
void USI_Stop_Timer( void ) {
	TCCR0B  = (0<<CS02)|(0<<CS01)|(0<<CS00); // Stop Timer
	USICR  =  0; 							 // Stop USI
}*/

void   timer_init( void )
{
	OCR1C = TIMER1_COMPC_MATCH;						  // OCR1C (immer) auf baudrate
	TCCR1B = (1<<PSR1) | (1<<CS10) | (1<<CS11);       // Set prescaler to /4 and start timer1
    TIFR   = (1<<TOV1);                               // Clear Timer1 OVF interrupt flag.
    //TIMSK |= (1<<TOIE1);         					  // Enable Timer1 OVF interrupt.
	return;
}

// void static inline USI_UART_start( void) {
 void USI_UART_start( void) {
// "inline" funktioniert nicht mit "Debug" Configuration (=ohne Optimierung)
// "Inline" spart aber 20 bytes

//	void   timer_init(); // nur neu aufrufen, wenn Gefahr besteht, dass der Timer abgestellt wird
	timer_init();
#if UART_DEBUG
			OCR1C = fudge; // allows to step through baudrates
#endif
			// OCR0A = 110; //104 + 6 empirischer Korrekturfaktor
//			TCCR1B = (1<<PSR1);								  // Reset the prescaler
//	        TCCR1A = (1<<WGM01);							  // Timer1 auf CTC setzen
//	        TCCR0B  = (0<<CS02)|(0<<CS01)|(1<<CS00);		  // Set prescaler to 1 and start Timer1.
			spiX_status.modeUART		 = 1;
	        TIMSK |= (1<<TOIE1);         					  // Enable Timer1 OVF interrupt.
	        USIPP  = (1<<USIPOS);							  // USI auf PORTA legen.
	        USICR  = (0<<USISIE)|(1<<USIOIE)|                 // Enable USI Counter OVF interrupt.
	                 (0<<USIWM1)|(1<<USIWM0)|                 // Select Three Wire mode.
//	                 (0<<USICS1)|(1<<USICS0)|(0<<USICLK)|     // Select Timer0 COMPA as USI Clock source.
					 (0<<USICS1)|(0<<USICS0)|(1<<USICLK)|	  // use software clock strobe
	                 (0<<USITC);
	                                                         // Note that enabling the USI will also disable the pin change interrupt.
	        return;
}

// Initialise USI for UART transmission.
void static inline USI_UART_Initialise_Transmitter( void )
{
    cli();
    TCNT1  = 0x00;
    USI_UART_start();
    USIDR  = 0xFF;                                  // Make sure MSB is '1' before enabling USI_DO.
    USISR  = 0xF0 |                                 // Clear all USI interrupt flags.
             0x0F;                                  // Preload the USI counter to generate interrupt at first USI clock.
    USI_UART_DDR  |= (1<<DO_TX);                      // Configure USI_DO as output.
                  
    USI_UART_status.ongoing_Transmission_From_Buffer = TRUE;

    sei();
    return;
}

// Initialise USI for UART reception.
// Note that this function only enables pinchange interrupt on the USI Data Input pin.
// The USI is configured to read data within the pinchange interrupt.
void inline USI_UART_Initialise_Receiver( void )
{  
    USI_UART_PORT |=  ((1<<DI_RX)|(1<<DO_TX));    // Enable pull up on USI DO and DI pins.
    USI_UART_DDR  &= ~((1<<DI_RX)|(1<<DO_TX));    // Set USI DI and DO pins as inputs.
    USICR  =  0;                         	// Disable USI.
    GIFR   =  (1<<PCIF);                 	// Clear pin change interrupt flag.
    PCMSK0 |=  (1<<DI_RX);					 	// Set pin change mask to DI pin
    GIMSK |=  (1<<PCIE1);                 	// Enable pin change interrupt.
    return;
}

// Puts data in the transmission buffer, after reverseing the bits in the byte.
// Initiates the transmission routines if not already started.
void USI_UART_Transmit_Byte( char data )
{
    uint8_t tmphead;

    tmphead = ( UART_TxHead + 1 ) & UART_TX_BUFFER_MASK;        // Calculate buffer index.
    while ( tmphead == UART_TxTail );                           // Wait for free space in buffer.
    UART_TxBuf[tmphead] = Bit_Reverse(data);                    // Reverse the order of the bits in the data byte and store data in buffer.
    UART_TxHead = tmphead;                                      // Store new index.
    
    if ( !USI_UART_status.ongoing_Transmission_From_Buffer )    // Start transmission from buffer (if not already started).
    {
        while ( USI_UART_status.ongoing_Reception_Of_Package ); // Wait for USI to finsh reading incoming data.
        USI_UART_Initialise_Transmitter();              
    }
    return;
}

void USI_UART_puts( char *s )			// transmit string from SRAM
{
  while( *s )
	  USI_UART_Transmit_Byte( *s++ );
  return;
}

void USI_UART_puts_p( const char *progmem_s )			// transmit string from PROGMEM
{
	register char c;
	while( (c = pgm_read_byte(progmem_s++)) )
	  USI_UART_Transmit_Byte( c );
	return;
}

// Returns a byte from the receive buffer. Waits if buffer is empty.
char USI_UART_Receive_Byte( void )
{
    uint8_t tmptail;
        
    while ( UART_RxHead == UART_RxTail );                 // Wait for incomming data <- blocking wait??
    tmptail = ( UART_RxTail + 1 ) & UART_RX_BUFFER_MASK;  // Calculate buffer index 
    UART_RxTail = tmptail;                                // Store new index 
    return Bit_Reverse(UART_RxBuf[tmptail]);              // Reverse the order of the bits in the data byte before it returns data from the buffer.
}


uint8_t USI_UART_Data_In_Transmit_Buffer( void )
{
    return ( UART_TxHead != UART_TxTail );                // Return 0 (FALSE) if the receive buffer is empty.
}

// Check if there is data in the receive buffer.
uint8_t USI_UART_Data_In_Receive_Buffer( void )
{
    return ( UART_RxHead != UART_RxTail );                // Return 0 (FALSE) if the receive buffer is empty.
}

// ********** Interrupt Handlers ********** //

// The pin change interrupt is used to detect USI_UART reception.
// It is here the USI is configured to sample the UART signal.

ISR(PCINT_vect)
{                                                                    
    if (!( PINA & (1<<DI_RX) ))                                     // If the USI DI pin is low, then it is likely that it
    {                                                             //  was this pin that generated the pin change interrupt.
    	TCNT1 =  INITIAL_TIMER1_SEED + INTERRUPT_STARTUP_DELAY; // Plant initial TIMER1 seed for first bit to match baudrate (incl interrupt start up time.).
    	//TCNT0 = 260; //206 bis 283 scheint zu funktionieren
    	USI_UART_start();
        USISR  = 0xF0 |                                           // Clear all USI interrupt flags.
                 USI_COUNTER_SEED_RECEIVE;                        // Preload the USI counter to generate interrupt.
                                                                  
        GIMSK &=  ~(1<<PCIE1);                                    // Disable pin change interrupt for PA7:0.
        // *** Besser DI pin aus PCMSK entfernen?
        USI_UART_status.ongoing_Reception_Of_Package = TRUE;             
    }
    return;
}

// The USI Counter Overflow interrupt is used for moving data between memmory and the USI data register.
// The interrupt is used for both transmission and reception.

ISR(USI_OVF_vect)
{
	if (spiX_status.modeUART == 1)
	{
		uint8_t tmphead,tmptail;

		// Check if we are running in Transmit mode.
		if( USI_UART_status.ongoing_Transmission_From_Buffer )
		{
			// If ongoing transmission, then send second half of transmit data.
			if( USI_UART_status.ongoing_Transmission_Of_Package )
			{
				USI_UART_status.ongoing_Transmission_Of_Package = FALSE;    // Clear on-going package transmission flag.

				USISR = 0xF0 | (USI_COUNTER_SEED_TRANSMIT);                 // Load USI Counter seed and clear all USI flags.
				USIDR = (USI_UART_TxData << 3) | 0x07;                      // Reload the USIDR with the rest of the data and a stop-bit.
			}
			// Else start sendinbg more data or leave transmit mode.
			else
			{
				// If there is data in the transmit buffer, then send first half of data.
				if ( UART_TxHead != UART_TxTail )
				{
					USI_UART_status.ongoing_Transmission_Of_Package = TRUE; // Set on-going package transmission flag.

					tmptail = ( UART_TxTail + 1 ) & UART_TX_BUFFER_MASK;    // Calculate buffer index.
					UART_TxTail = tmptail;                                  // Store new index.
					USI_UART_TxData = UART_TxBuf[tmptail];                  // Read out the data that is to be sent. Note that the data must be bit reversed before sent.
																			// The bit reversing is moved to the application section to save time within the interrupt.
					USISR  = 0xF0 | (USI_COUNTER_SEED_TRANSMIT);            // Load USI Counter seed and clear all USI flags.
					USIDR  = (USI_UART_TxData >> 2) | 0x80;                 // Copy (initial high state,) start-bit and 6 LSB of original data (6 MSB
																			//  of bit of bit reversed data).
				}
				// Else enter receive mode.
				else
				{
					USI_UART_status.ongoing_Transmission_From_Buffer = FALSE;
					USI_UART_Initialise_Receiver();
				}
			}
		}

		// Else already running in receive mode.
		else
		{
			USI_UART_status.ongoing_Reception_Of_Package = FALSE;

			tmphead     = ( UART_RxHead + 1 ) & UART_RX_BUFFER_MASK;        // Calculate buffer index.

			if ( tmphead == UART_RxTail )                                   // If buffer is full trash data and set buffer full flag.
			{
				USI_UART_status.reception_Buffer_Overflow = TRUE;           // Store status to take actions elsewhere in the application code
			}
			else                                                            // If there is space in the buffer then store the data.
			{
				UART_RxHead = tmphead;                                      // Store new index.
				UART_RxBuf[tmphead] = USIBR;                                // Store received data in buffer. Note that the data must be bit reversed before used.
			}                                                               // The bit reversing is moved to the application section to save time within the interrupt.

			USI_UART_Initialise_Receiver();
		}
	}
	else
	{
		/* SPI Mode:
		*  Master must now disable the overflow interrupt
		* to prevent more USI counter clocks.
		*/
		if( spiX_status.masterMode == 1 ) {
			TIMSK &= ~(1<<TOIE1);
			// SPI_TIMER_INTERRUPT_REG &= ~(1<<SPI_TIMER_INTERRUPT);
		}

		// Update flags and clear USI counter
		USISR = (1<<USIOIF);
		spiX_status.transferComplete = 1;

		// Copy USIDR to buffer to prevent overwrite on next transfer.
		storedUSIDR = USIDR; // Todo: besser USIBR lesen? Ausprobieren...
		return;
	}
	return;

}

// Timer1 OVF interrupt is used to trigger the sampling of signals on the USI ports.

ISR(TIMER1_OVF_vect)
{
	if (spiX_status.modeUART == 1)
		USICR  |= (1<<USICLK);  // TODO: Was passiert, wenn man dieselbe clocksource verwendet wie bei SPI?
	else
		USICR |= (1<<USITC);	// Toggle clock output pin.
	return;
}
